/*
 * Demos.cpp - Ports of the ModernDevice.com Demos for the Fluxmasynth Shield
 *
 * Port by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "LiquidCrystal.h"
#include "SAM2195.h"
#include "Utils.h"

extern LiquidCrystal lcd;

// All voices demo - Plays all voices with a random algorithmic tune
// This demo will run until right joystick activity is detected
boolean allVoicesDemoFunction(boolean firstTIme) {

    // Do demo setup
    int counter = 0;
    int channel = 0;
    int voice = 0;
    int j = 0;

    // Display status line on lcd
    lcd.setCursor(0, 1);
    lcd.print("Demo 1 Running");

    programChange(0, 9, 40);
    programChange(0, 1,  0);

    // Run until right joystick is detected
    while (true) {

        while (counter < 128) {
            if (voice > 128) {
                voice = 0;
            }
            programChange(channel, 0, voice);

            for (int i = 0; i < 10; i++) {
                int note = random(36,100);
                noteOn(channel, note, 127);
                int coinflip = random(0, 20);
                if (coinflip == 5) {
                    while (j < 200) {
                        pitchBend(channel, j);
                        j+= 16;
                        delay(10);
                    }
                    j = 0;
                }
                // Was the right joystick detected ?
                if (pollRightJoystick()) {
                    midiReset();
                    return true;
                }
                delay(100);
                noteOff(channel, note);
            }
            voice++;
        }
        channel++;
        counter++;
    }
}

// Note bending demo - Demonstrates note bending with the SAM2195
// This demo will run until right joystick activity is detected
boolean noteBendingDemoFunction(boolean firstTime) {

    // Display status line on lcd
    lcd.setCursor(0, 1);
    lcd.print("Demo 2 Running");

    // Do demo setup
    int i = 0x2000;          // Middle of pitch bend range

    programChange(127, 0, 91);
    pitchBendRange(0, 4);    // Double the pitch bend sensitivity (4 semitones)

    noteOn(0, 62, 127);

    while (true) {
        while (i < 0x3FFF) {   // Loop to upper bound of pitch bend range
            pitchBend(0, i);
            i += 32;
        }

        if (pollRightJoystick()) {
            midiReset();
            return true;
        }

        while (i > 0) {      // Loop to lower bound of pitch bend range
            pitchBend(0, i);
            i -= 32;
        }
    }
}

// Cross fading demo - Demonstrates cross channel fading with the SAM2195
// This demo will run until right joystick activity is detected
boolean crossFadeDemoFunction(boolean firstTime) {

    // Display status line on lcd
    lcd.setCursor(0, 1);
    lcd.print("Demo 3 Running");

    // Note to use    
    int note = 60;

    // Complementary volumes
    int vol1 = 0;                  // Volumes are inverses of each other
    int vol2 = 127 - vol1;

    int stepSize = 1;              // Step size
    int dir = 1;                   // Direction to step in: 1 = up, -1 = down

    programChange(0, 0,  42);      // Give our two channels different voices
    programChange(0, 1,  88);

    setChannelVolume(0, vol1);     // Set their initial volumes
    setChannelVolume(1, vol2);

    noteOn(0, note, 127);         // Turn on a note for each channel
    noteOn(1, note, 127);

    // Loop until right joystick is manipulated
    while (! pollRightJoystick()) {
        while ((vol1 < 128) && (vol1 >= 0)) { // Loop while in valid volume range
            setChannelVolume(0, vol1); // Update volumes
            setChannelVolume(1, vol2);
            vol1 += stepSize * dir;    // vol1 increments by step size in direction dir
            vol2 = 127 - vol1;         // vol2 is always the inverse
            delay(20);
        }
        dir *= -1;                     // Once volumes reach bounds, reverse dir
        vol1 += stepSize * dir;        // Increment once in new direction
        vol2 = 127 - vol1;             // to ensure volumes are back in range
    }
    midiReset();
    return true;
}

// Define midi note numbers for several GM drum sounds
#define bass   36              
#define snare  38
#define hihatC 42
#define hihatP 44
#define hihatO 46

// The drum pattern
int patternSize = 16; // A pattern of 16 ticks (actually 15 because tick 16 is tick 1)

// Every array cell is the velocity of the note played 
// Tick         1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
int bd  [] = {
    127,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 90,  0}; // Bass drum
int sn  [] = {  
    0,  0,  0,  0,127,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};   // Snare
int hho [] = {  
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,127,  0};   // Hihat Open
int hhc [] = {
    127, 40, 80, 40,127, 40, 80, 40,127, 40, 80, 40,127,  0,  0}; // Hihat Close
int hhp [] = {  
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,127};   // Hihat Pedal

// Play bass drum
void bassDrum (int channel, int vel) {
    noteOn(channel, bass, vel); // Send note on
    noteOff(channel, bass);     // Send note off
}

// Play snare drum
void snareDrum (int channel, int vel) {
    noteOn(channel, snare, vel);  // Send note on
    noteOff(channel, snare);      // Send note off
}

// Play hi hat closed
void hihatClose (int channel, int vel) {
    noteOn(channel, hihatC, vel); // Send note on
    noteOff(channel, hihatC);     // Send note off 
}

// Play hi hat pedal
void hihatPedal (int channel, int vel) {
    noteOn(channel, hihatP, vel); // Send note on
    noteOff(channel, hihatP);     // Send note off   
}

// Play hi hat open
void hihatOpen (int channel, int vel) {
    noteOn(channel, hihatO, vel); // Send note on
    noteOff(channel, hihatO);     // Send note off 
}

// Drum pattern demo - Demonstrates drum sounds with the SAM2195
// This demo will run until right joystick activity is detected
boolean drumPatternDemoFunction(boolean firstTime) {

    // Display status line on lcd
    lcd.setCursor(0, 1);
    lcd.print("Demo 4 Running");

    // Basic settings
    int channel = 9;    // MIDI channel number
    int tempo = 120;    // Tempo

    // Program 
    // 0: Room1   1: Room2    2: Room3 
    // 3: Hall1   4: Hall2    5: Plate
    // 6: Delay   7: Pan delay
    setReverb(channel, 5, 255, 100); // A Plate Reverb with maximum effect level

    // Run until joystick activity is detected, then quit
    while (! pollRightJoystick()) {
        for(int tickNo = 0; tickNo < patternSize; tickNo++) {
            bassDrum(channel,   bd[tickNo]);
            snareDrum(channel,  sn[tickNo]);
            hihatClose(channel, hhc[tickNo]);
            hihatOpen(channel,  hho[tickNo]);
            hihatPedal(channel, hhp[tickNo]);

            delay (tempo);
        }
    }
    midiReset();
    return true;
}


















