/*
 * MIDI Buddy - MIDI Synthesizer Project using Teensy 3.1 and the SAM2195 MIDI/Synth chip
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "Demos.h"
#include "FileFunctions.h"
#include "LiquidCrystal.h"
#include "SdFat.h"
#include "SAM2195.h"
#include "AUISMFunctions.h"
#include "Utils.h"

// LCD connections
#define LCD_RS_PIN        2
#define LCD_E_PIN         3
#define LCD_DB4_PIN       4
#define LCD_DB5_PIN       5
#define LCD_DB6_PIN       6
#define LCD_DB7_PIN       7
#define LCD_BACKLIGHT_PIN 8

// Name of this application
#define APPLICATION_NAME "   MIDI Buddy   "

// MIDI port definitions
#define MIDI_IN_PORT_NUMBER  1
#define MIDI_OUT_PORT_NUMBER 2

// SD card chip select
#define SD_CARD_CS_PIN   14

// Define hardware driver instances
SdFat sd;
SdFile file;

// Instance of lcd object
LiquidCrystal lcd(LCD_RS_PIN, LCD_E_PIN, LCD_DB4_PIN, LCD_DB5_PIN, LCD_DB6_PIN, LCD_DB7_PIN);

// External variable definitions
extern boolean recording;
extern void runRecordingFSM(int data);

/*******************************************************************/
/***                   Named Menu Definitions                    ***/
/*******************************************************************/

// Pointer to a function taking one parameter and returning boolean
typedef boolean (*ptr2Function)(boolean firstTime);

// Mode definition structure
typedef struct {
    char *menuName;
    ptr2Function menuFunction;
} 
NAMED_FUNCTION;

// Each items in this array represents a single menu item
NAMED_FUNCTION menuItems [] = {
    "Play MIDI File",  playMidiFileFunction,
    "Record Riff",     recordRiffFunction,
    "Play Riff",       playRiffFunction,
    "Master Volume",   masterVolumeFunction,
    "Drum Set",        drumSetFunction,
    "Synth Reset",     resetFunction,
    "Chan Volume",     channelVolumeFunction,
    "Chan Program",    channelProgramFunction,
    "Chan Pan",        channelPanFunction,
    "Chan Chorus",     channelChorusFunction,
    "Chan Reverb",     channelReverbFunction,
    "Chan Portamento", channelPortamentoFunction,
    "Chan Modulation", channelModulationFunction,
    "D1: All Voices",  allVoicesDemoFunction,
    "D2: Note Bending",noteBendingDemoFunction,
    "D3: Cross Fade",  crossFadeDemoFunction,
    "D4: Drum Pattern",drumPatternDemoFunction,
};

#define NUMBER_OF_MENU_ITEMS (sizeof(menuItems) / sizeof(NAMED_FUNCTION))

/*******************************************************************/
/***                 High Level UI State Machine                 ***/
/*******************************************************************/

#define LOOP_COUNT 2000000

// Possible UI states
enum UI_STATES {
    STATE_BANNER, STATE_DELAY, STATE_DISPLAY, STATE_POLLING, STATE_READING, STATE_DISPATCHING};

enum UI_STATES uiState = STATE_BANNER;
int menuIndex;
boolean firstTime;
static int count;
enum JS_DIRECTION direction;
ptr2Function menuFunction;

void runUIStateMachine() {
    int result;

    switch(uiState) {

    case STATE_BANNER:    // Display current menu selection
        lcd.clear();
        lcd.print(APPLICATION_NAME);

        count = LOOP_COUNT;
        uiState = STATE_DELAY;
        return;

    case STATE_DELAY:
        count--;
        if (count <= 0) {
            uiState = STATE_DISPLAY;
        }
        return;

    case STATE_DISPLAY:    // Display current menu selection
        lcd.clear();
        lcd.print(menuItems[menuIndex].menuName);

        firstTime = true;

        // Advance to next state
        uiState = STATE_POLLING;
        return;

    case STATE_POLLING:
        // Is there left joystick activity?
        if (pollLeftJoystick()) {
            // Advance to next state
            uiState = STATE_READING;
        }
        return;

    case STATE_READING:
        // Read the left joystick
        direction = readLeftJoystick();

        switch (direction) {

        case DIR_UP:
            menuIndex--;
            if (menuIndex < 0) {
                menuIndex = NUMBER_OF_MENU_ITEMS - 1;
            }
            uiState = STATE_DISPLAY;
            return;

        case DIR_DOWN:
            menuIndex++;
            if (menuIndex >= NUMBER_OF_MENU_ITEMS) {
                menuIndex = 0;
            }
            uiState = STATE_DISPLAY;
            return;

        case DIR_RIGHT:
            // User has selected menu item
            uiState = STATE_DISPATCHING;
            return;

        case DIR_LEFT:
            uiState = STATE_POLLING;
            return;
        }
        break;

    case STATE_DISPATCHING:
        // Get the selected menu function
        menuFunction = menuItems[menuIndex].menuFunction;

        // Run the function. If true is returned, function has completed
        if ((* menuFunction)(firstTime)) {
            uiState = STATE_BANNER;
        }
        firstTime = false;
        return;
    }
}

/*******************************************************************/
/***                     Main Program Setup                     ***/
/*******************************************************************/

void setup() {

    // Setup serial interface
    Serial.begin(250000);

    Serial1.begin(31250);

    // Wait for serial interface to settle
    delay(2000);

    // Seed the random number generator
    randomSeed(analogRead(A14));

    // Set the logging level
    setLogLevel(LOG_WARNING);

    // Initialize SD card interface
    logger(LOG_MSG, "Initializing SD card\n");
    if (! sd.begin(SD_CARD_CS_PIN, SPI_HALF_SPEED)) {
        sd.initErrorHalt();
    }
    logger(LOG_MSG, "SD card initialized\n");

    // Setup joystick switches with active pullup
    // Joystick switches are not currently used
    pinMode(JOYSTICK_LEFT_BUTTON, INPUT);
    digitalWrite(JOYSTICK_LEFT_BUTTON, HIGH);
    pinMode(JOYSTICK_RIGHT_BUTTON, INPUT);
    digitalWrite(JOYSTICK_RIGHT_BUTTON, HIGH);

    // Setup for control of LCD backlight
    pinMode(LCD_BACKLIGHT_PIN, OUTPUT);
    digitalWrite(LCD_BACKLIGHT_PIN, HIGH); 

    // set up the LCD's number of columns and rows: 
    lcd.begin(16, 2);

    // Print app name message to the LCD.
    lcd.print(APPLICATION_NAME);

    delay(1000);

    // Prepare MIDI output port
    selectAndInitializeSerialPort(MIDI_OUT_PORT_NUMBER);

    // Do a complete reset of SAM2195 synth
    midiReset();

    // Initialize the MIDI environment
    initializeMIDIEnv();

    // Enumerate all of the MIDI files on SD card
    enumerateMIDIFiles(false);

    // We are not recording
    recording = false;
}

/*******************************************************************/
/***                      Main Program Loop                      ***/
/*******************************************************************/

void loop() {

    int data;

    while (true) {
        // Read the MIDI in data and write it to the synth through MIDI_OUT_PORT_NUMBER
        while (Serial1.available()) {

            // Read the MIDI data at the input port
            data = Serial1.read();

            // Are we recording ?
            if (recording) {
                // Run recording FSM
                runRecordingFSM(data);
            }
            // MIDI data filter functions would go here
            // Write (possibily modified) data to the output port
            write(data);
        }

        // Run the main UI state machine
        runUIStateMachine();
    }
}


