/*
 * MIDIParser.h - MIDI Parser Header File
 *
 * Standard Type 0 MIDI File Parser Code
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "Arduino.h"

#ifndef MIDIPARSER_H
#define MIDIPARSER_H

// Error codes that can occur when parsing a Type 0 MIDI file
#define ERROR_NONE            0
#define ERROR_FILEOPEN       -1
#define ERROR_FILENOTSMF     -2
#define ERROR_FILENOTTYPE0   -3
#define ERROR_BADFORMAT      -4
#define ERROR_TIMEDIVISION   -5

// Special callback for reporting timeDivision for SMF header
typedef void (*timeDivisionCallbackType)(int ppqn);

// Callbacks for MIDI events
typedef void (*noteOffCallbackType)(uint32_t tics, int channel, int note, int velocity);
typedef void (*noteOnCallbackType)(uint32_t tics, int channel, int note, int velocity);
typedef void (*noteAfterTouchCallbackType)(uint32_t tics, int channel, int note, int value);
typedef void (*controllerCallbackType)(uint32_t tics, int channel, int controller, int value);
typedef void (*programChangeCallbackType)(uint32_t tics, int channel, int program);
typedef void (*channelAfterTouchCallbackType)(uint32_t tics, int channel, int value);
typedef void (*pitchBendCallbackType)(uint32_t tics, int channel, int value);

// Callbacks for META events
typedef void (*sequenceNumberCallbackType)(uint32_t tics, int numberMSB, int numberLSB);
typedef void (*textEventCallbackType)(uint32_t tics, char *text);
typedef void (*copyrightNoticeCallbackType)(uint32_t tics, char *text);
typedef void (*sequenceTrackNameCallbackType)(uint32_t tics, char *text);
typedef void (*instrumentNameCallbackType)(uint32_t tics, char *text);
typedef void (*lyricCallbackType)(uint32_t tics, char *text);
typedef void (*markerCallbackType)(uint32_t tics, char *text);
typedef void (*cuePointCallbackType)(uint32_t tics, char *text);
typedef void (*midiChannelPrefixCallbackType)(uint32_t tics, int channel);
typedef void (*setTempoCallbackType)(uint32_t tics, int tempoUSPQN);
typedef void (*smpteOffsetCallbackType)(uint32_t tics, int hours, int mins, int secs, int frames, int subFrames);
typedef void (*timeSignatureCallbackType)(uint32_t tics, int numerator, int denominator, int metro, int _32nds);
typedef void (*keySignatureCallbackType)(uint32_t tics, int key, int scale);
typedef void (*sequencerSpecificMetaEventCallbackType)(uint32_t tics, char *data);

void setHandleTimeDivisionCallback(timeDivisionCallbackType func); 

// Functions to set MIDI event callbacks
void setHandleNoteOffCallback(noteOffCallbackType func);   
void setHandleNoteOnCallback(noteOnCallbackType func);   
void setHandleNoteAfterTouchCallback(noteAfterTouchCallbackType func);    
void setHandleControllerCallback(controllerCallbackType func);
void setHandleProgramChangeCallback(programChangeCallbackType func);    
void setHandleChannelAfterTouchCallback(channelAfterTouchCallbackType func);
void setHandlePitchBendCallback(pitchBendCallbackType func);

// Functions to set META event callbacks
void setHandleSequenceNumberCallback(sequenceNumberCallbackType func);
void setHandleTextEventCallback(textEventCallbackType func);
void setHandleCopyrightNoticeCallback(copyrightNoticeCallbackType func);
void setHandleSequenceTrackNameCallback(sequenceTrackNameCallbackType func);    
void setHandleInstrumentNameCallback(instrumentNameCallbackType func);
void setHandleLyricCallback(lyricCallbackType func);
void setHandleMarkerCallback(markerCallbackType func);
void setHandleCuePointCallback(cuePointCallbackType func);
void setHandleMidiChannelPrefixCallback(midiChannelPrefixCallbackType func);
void setHandleSetTempoCallback(setTempoCallbackType func);
void setHandleSMPTEOffsetCallback(smpteOffsetCallbackType func);
void setHandleTimeSignatureCallback(timeSignatureCallbackType func);
void setHandleKeySignatureCallback(keySignatureCallbackType func);
void setHandleSequencerSpecificMetaEventCallback(sequencerSpecificMetaEventCallbackType func);

void terminateMIDIPlayback();

// Function to parse a Type 0 SMF
int parseSMF (const char *pathname);

#endif

