/*
 * MIDIPlayer.cpp - MIDI File Player Code
 *
 * Plays standard Type 0 MIDI Files
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "MIDIParser.h"
#include "SAM2195.h"
#include "Utils.h"

int bpm;
int ppqn;
int ticTimeInMicroseconds;
uint32_t ticCount;

const uint32_t MICROSECONDS_PER_MINUTE = 60000000;

// Calculate BPM from MSPQN
int calculateBPMfromMSPQN(uint32_t mspqn) {

    return MICROSECONDS_PER_MINUTE / mspqn;
}

void calculateTicTimeInMicroseconds(int bpm, int ppqn) {

    ticTimeInMicroseconds = MICROSECONDS_PER_MINUTE / (bpm * ppqn);
}

// Wait until the ticCount for the current event occurs
void waitForTicCount(int targetTicCount) {

    while (true) {
        if (targetTicCount < ticCount) {
            return;
        }
        delayMicroseconds(ticTimeInMicroseconds);
        ticCount++;
    }
}

// Function for reporting time division in PPQN from SMF header
void handleTimeDivision(int _ppqn) {

    logger(LOG_INFO, "Time Division (PPQN): %d\n", _ppqn);
    ppqn = _ppqn;

    // Now calculate ticTime
    calculateTicTimeInMicroseconds(bpm, ppqn);
    logger(LOG_INFO, "ticTime in microseconds: %d\n", ticTimeInMicroseconds);
}

// Functions for handling midi events and meta events
void handleNoteOff(uint32_t tics, int channel, int note, int velocity) {

    logger(LOG_INFO, "Note Off at: %d for Channel: %d, Note: %d, Velocity: %d\n", tics, channel, note, velocity);

    waitForTicCount(tics);
    noteOff(channel, note);
}

void handleNoteOn(uint32_t tics, int channel, int note, int velocity) {

    logger(LOG_INFO, "Note On at: %d for Channel: %d, Note: %d, Velocity: %d\n", tics, channel, note, velocity);

    waitForTicCount(tics);
    noteOn(channel, note, velocity);
}

void handleNoteAfterTouch(uint32_t tics, int channel, int note, int value) {

    logger(LOG_INFO, "Note After Touch at: %d for Channel: %d, Note: %d, Value: %d\n", tics, channel, note, value);

}

void handleController(uint32_t tics, int channel, int controller, int value) {

    logger(LOG_INFO, "Controller at: %d for Channel: %d, Controller: %d, Value: %d\n", tics, channel, controller, value);

    waitForTicCount(tics);
    controllerChange(channel, controller, value);
}

void handleProgramChange(uint32_t tics, int channel, int program) {

    logger(LOG_INFO, "Program Change at: %d for Channel: %d, Program: %d\n", tics, channel, program);

    waitForTicCount(tics);
    programChange(0, channel, program);
}

void handleChannelAfterTouch(uint32_t tics, int channel, int value) {

    logger(LOG_INFO, "Channel After Touch at: %d for Channel: %d, Value: %d\n", tics, channel, value);
}

void handlePitchBend(uint32_t tics, int channel, int value) {

    logger(LOG_INFO, "Pitch Bend at: %d for Channel: %d, Value: %d\n", tics, channel, value);
    
    pitchBend(channel, value);
}

void handleSequenceNumber(uint32_t tics, int numberMSB, int numberLSB) {

    logger(LOG_INFO, "Sequence at: %d - NumberMSB: %d, NumberLSB: %d\n", tics, numberMSB, numberLSB);
}

void handleTextEvent(uint32_t tics, char *text) {

    logger(LOG_INFO, "Text Eventat: %d: - %s\n", tics, text);
}

void handleCopyrightNotice(uint32_t tics, char *text) {

    logger(LOG_INFO, "Copyright Notice at: %d: - %s\n", tics, text);
}

void handleSequenceTrackName(uint32_t tics, char *text) {

    logger(LOG_INFO, "Sequence Track Name at: %d: - %s\n", tics, text);
}

void handleInstrumentName(uint32_t tics, char *text) {

    logger(LOG_INFO, "Instrument Name at: %d: - %s\n", tics, text);
}

void handleLyric(uint32_t tics, char *text) {

    logger(LOG_INFO, "Lyric at: %d: - %s\n", tics, text);
}

void handleMarker(uint32_t tics, char *text) {

    logger(LOG_INFO, "Marker at: %d: - %s\n", tics, text);
}

void handleCuePoint(uint32_t tics, char *text) {

    logger(LOG_INFO, "Cue Point at: %d: - %s\n", tics, text);
}

void handleMidiChannelPrefix(uint32_t tics, int channel) {

    logger(LOG_INFO, "Midi Channel Prefix at: %d: - %d\n", tics, channel);
}

void handleSetTempo(uint32_t tics, int tempo) {

    logger(LOG_INFO, "Set Tempo (micro seconds / quarter note) at: %d - %d\n", tics, tempo);

    bpm = calculateBPMfromMSPQN(tempo);
    
    // Now calculate ticTime
    calculateTicTimeInMicroseconds(bpm, ppqn);
    logger(LOG_INFO, "ticTime in microseconds: %d\n", ticTimeInMicroseconds);
}

void handleSMPTEOffset(uint32_t tics, int hours, int mins, int secs, int frames, int subFrames) {

    logger(LOG_INFO, "SMPTE Offset at: %d - Hours: %d, Mins: %d, Secs: %d, Frames: %d, Sub Frames: %d\n", 
    tics, hours, mins, secs, frames, subFrames);
}

void handleTimeSignature(uint32_t tics, int numerator, int denominator, int metro, int _32nds) {

    logger(LOG_INFO, "Time Signature at: %d - Numerator: %d, Denominator: %d, Metro: %d, 32nds: %d\n",
    tics, numerator, denominator, metro, _32nds); 
}

void handleKeySignature(uint32_t tics, int key, int scale) {

    logger(LOG_INFO, "Key Signature at: %d - Key: %d, Scale: %d\n", tics, key, scale);
}

void handleSequencerSpecificMetaEvent(uint32_t tics, char *data) {

    logger(LOG_INFO, "Sequencer Specific Meta Event at: %d - Data: \n", tics, data);
}

// Play a Type 0 SMF 
int playSMF (const char *pathname) {

    // Install handlers for the various events generated by MIDIParser
    setHandleTimeDivisionCallback(handleTimeDivision);

    // Set handlers for MIDI events
    setHandleNoteOffCallback(handleNoteOff);   
    setHandleNoteOnCallback(handleNoteOn);   
    setHandleNoteAfterTouchCallback(handleNoteAfterTouch);    
    setHandleControllerCallback(handleController);
    setHandleProgramChangeCallback(handleProgramChange);    
    setHandleChannelAfterTouchCallback(handleChannelAfterTouch);
    setHandlePitchBendCallback(handlePitchBend);

    // Set handlers for META events
    setHandleSequenceNumberCallback(handleSequenceNumber);
    setHandleTextEventCallback(handleTextEvent);
    setHandleCopyrightNoticeCallback(handleCopyrightNotice);
    setHandleSequenceTrackNameCallback(handleSequenceTrackName);    
    setHandleInstrumentNameCallback(handleInstrumentName);
    setHandleLyricCallback(handleLyric);
    setHandleMarkerCallback(handleMarker);
    setHandleCuePointCallback(handleCuePoint);
    setHandleMidiChannelPrefixCallback(handleMidiChannelPrefix);
    setHandleSetTempoCallback(handleSetTempo);
    setHandleSMPTEOffsetCallback(handleSMPTEOffset);
    setHandleTimeSignatureCallback(handleTimeSignature);
    setHandleKeySignatureCallback(handleKeySignature);
    setHandleSequencerSpecificMetaEventCallback(handleSequencerSpecificMetaEvent);

    // Initialize default BPM
    bpm = 120;

    // Initialize tic counter to zero
    ticCount = 0;

    // Call the MIDI parser with the path/file name
    return parseSMF(pathname);
}









