/*
 * SAM2195.cpp - Interface Functions for SAM2195 MIDI Synth Chip
 *
 * Based upon Modern Device Fluxamasynth Arduino Library
 * but using hardware UART instead of NewSoftSerial.
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "SAM2195.h"
#include "Utils.h"

HardwareSerial *port;
boolean initialized = false;

void selectAndInitializeSerialPort (int portNumber) {

    // Only need to do initialization once
    if (initialized) {
        return;
    }

    switch (portNumber) {

    case 1:
        port = &Serial1;
        break;

    case 2:
        port = &Serial2;
        break;

    case 3:
        port = &Serial3;
        break;

    default:
        Serial.print("Bad serial port number specified: ");
        Serial.println(portNumber);
        return;

    }
    port->begin(31250);
    delay(20);
    initialized = true;
}

// Write a byte to the serial port
int write (byte b) {
    if (initialized) {
        return port->write(b);
    }
    return -1;    
}

// Write a buffer to the serial port
int write (byte *buffer, int count) {
    if (initialized) {
        return port->write(buffer, count);
    }
    return -1;    
}

// Send a control change message
void controllerChange(byte channel, byte controlNumber, byte value) {
    byte command[3] = {
        0xB0 | (channel & 0x0F), (controlNumber & 0x7F), (value & 0x7F)                    };
    write(command, 3);
}

// Send a registered parameter number control message
void RPN_Control(byte channel, byte rpnHi, byte rpnLo, byte value) {
    // Data is of the form: [Bx] 65 hi 64 lo 06 vh
    byte command[7] = {
        0xB0 | (channel & 0x0F), 0x65, (rpnHi & 0x7f), 0x64, (rpnLo & 0x7f), 0x06, (value & 0x7F)                                };
    write(command, 7);
}

// Send a non registered parameter number control message
void NRPN_Control(byte channel, byte nrpnHi, byte nrpnLo, byte value) {
    // Data is of the form: [Bx] 63 hi 62 lo 06 val
    byte command[7] = {
        0xB0 | (channel & 0x0F), 0x63, (nrpnHi & 0x7f), 0x62, (nrpnLo & 0x7f), 0x06, (value & 0x7F)                                };
    write(command, 7);
}

// Send a non standard MIDI Dream control messaage
void DREAM_Control(byte functionNumber, byte value) {
    NRPN_Control(0x00, 0x37, functionNumber, value);
}

void sendGSParameterData(byte *data, int length) {
    // Send a Parameter Control sysex message.
    // All but two of the sysex blocks that Atmel Dream 2195 respond to
    // use a common Roland GS compatible header, so it's hard coded here.
    byte header [6] = {
        0xF0, 0x41, 0x00, 0x42, 0x12, 0x40        };

    // Checksum is ignored by 2195 so we don't transmit it
    byte trailer [1] = {
        0xF7        };

    write(header, 6);
    write(data, length);
    write(trailer, 2);
}

// Write a noteOn message
void noteOn (byte channel, byte pitch, byte velocity) {
    byte command[3] = {
        0x90 | (channel & 0x0F), pitch, velocity        };
    write(command, 3);
}

// Write a noteOff message
void noteOff (byte channel, byte pitch) {
    byte command[3] = { 
        0x80 | (channel & 0x0F), pitch, 0x00        };
    write(command, 3);
}

// Program change message
void programChange (byte bank, byte channel, byte value) {
    // Do control change
    controllerChange(channel, 0x00, bank);

    write(0xC0 | (channel & 0x0F));
    write(value);
}

// Drumset change message
void drumSetChange (byte set) {

    // Drums are channel 10 or 0x09
    write(0xC0 | 0x09);
    write(set);
}

// Pitch bend message
void pitchBend (byte channel, int value) {
    byte command[3] = { 
        0xE0 | (channel & 0x0F), (value & 0x7F), ((value >> 7) & 0x7F)                    };
    write(command, 3);
}

// Pitch bend range or sensitivity message
void pitchBendRange (byte channel, byte value) {
    RPN_Control(channel, 0x00, 0x00, value);
}

// MIDI reset message. Resets and enables standard MIDI
void midiReset() {
    write(0xFF);
}

// General MIDI reset message. Resets and enables GM MIDI
void gmReset() {
    // Data is of the form: F0H 7EH 7FH 09H 01H F7H
    byte command[6] = {
        0xF0, 0x7E, 0x7F, 0x09, 0x01, 0xF7                    };
    write(command, 6);
}

// Roland GS or General Standard MIDI reset message. Resets and enables GS MIDI
void gsReset() {
    byte data [3] = {
        0x00, 0x7F, 0x00                    };
    sendGSParameterData(data, 3);
}

// Channel volume message
void setChannelVolume (byte channel, byte level) {
    // Do control change
    controllerChange(channel, 0x07, level);
}

void setChannelPan (int channel, int value) {
    // Do control change
    controllerChange(channel, 0x0A, value);
}


// All notes off message
void allNotesOff (byte channel) {
    // Do control change
    controllerChange(channel, 0x7B, 0);
}

// Master volume message
void setMasterVolume (byte level) {
    //F0H 7FH 7FH 04H 01H 00H ll F7H
    byte command[8] = { 
        0xF0, 0x7F, 0x7F, 0x04, 0x01, 0x00, (level & 0x7F), 0xF7                                             };
    write(command, 8);
}

// Set reverb send message
void setReverb (byte channel, byte program, byte level, byte feedback) {
    // Change reverb program
    controllerChange(channel, 0x50, program & 0x07);

    // Change reverb send level
    controllerChange(channel, 0x5B, level);

    if (feedback > 0) {
        byte data [3] = {
            0x01, 0x35, (feedback & 0x7F)                                        };
        sendGSParameterData(data, 3);
    }
}

// Set chorus send message
void setChorus (byte channel, byte program, byte level, byte feedback, byte chorusDelay) {

    // Change chorus program
    controllerChange(channel, 0x51, program & 0x07);

    // Change chorus send level
    controllerChange(channel, 0x5D, level);

    if (feedback > 0) {
        byte data [3] = {
            0x01, 0x3B, (feedback & 0x7F)
            };
            sendGSParameterData(data, 3);
    }

    if (chorusDelay > 0) {
        byte data [3] = {
            0x01, 0x3C, (chorusDelay & 0x7F)
            };
            sendGSParameterData(data, 3);
    }
}

// Set modulation 
void setModulation(byte channel, byte depth, byte rate) {

    // Change modulation depth
    controllerChange(channel, 0x01, depth);

    byte data [3] = {0x20, 0x03, (rate & 0x7F)};
    sendGSParameterData(data, 3);
}

// Set portamento
void setPortamento(byte channel, boolean on, byte time) {

    // Change portameno state
    controllerChange(channel, 0x41, ((on) ? 0x7F : 0x00));

    // Change portamento time
    controllerChange(channel, 0x05, time);
}




