/*
 * UISMFunctions.cpp - UI state machine functions for the MIDI Synthesizer Project
 *
 * Written by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 09/10/2014
 *
 * Copyright (c) 2014 Craig A. Lindley
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "AUISMFunctions.h"
#include "FileFunctions.h"
#include "LiquidCrystal.h"
#include "MIDIInfo.h"
#include "MIDIPlayer.h"
#include "SAM2195.h"
#include "SdFat.h"

#include "Utils.h"

extern LiquidCrystal lcd;
extern SdFat sd;

#define DEFAULT_VOLUME_LEVEL 100
#define DEFAULT_PAN_LEVEL     64
#define DEFAULT_CHORUS_PROGRAM 2
#define DEFAULT_REVERB_PROGRAM 4
#define VOLUME_INCREMENT (128 / 16)

#define LOOP_COUNT 2000000

struct CHANNEL_RECORD {

    // Channel parameters
    int16_t bank;
    int16_t program;
    int16_t volume;
    int16_t pan;

    // Chorus parameters
    int16_t chorusProgram;
    int16_t chorusLevel;
    int16_t chorusFeedback;
    int16_t chorusDelay;

    // Reverb parameters
    int16_t reverbProgram;
    int16_t reverbLevel;
    int16_t reverbFeedback;

    // Portamento parameters
    int16_t portamentoState;
    int16_t portamentoTime;

    // Modulation parameters
    int16_t modulationDepth;
    int16_t modulationRate;
};

// MIDI environment
struct MIDI_ENVIRONMENT {

    int16_t masterVolume;
    int16_t fileIndex;
    int16_t drumSetIndex;

    // Array of channel records; one for each channel
    struct CHANNEL_RECORD channels[NUMBER_OF_MIDI_CHANNELS];
} 
MIDI_ENV;

// Initialize the MIDI environment
void initializeMIDIEnv() {

    // Set master volume
    MIDI_ENV.masterVolume = DEFAULT_VOLUME_LEVEL;
    setMasterVolume(DEFAULT_VOLUME_LEVEL);

    MIDI_ENV.fileIndex = 0;
    MIDI_ENV.drumSetIndex = 0;

    for (int ch = 0; ch < NUMBER_OF_MIDI_CHANNELS; ch++) {

        MIDI_ENV.channels[ch].bank = 0;
        MIDI_ENV.channels[ch].program = 0;
        MIDI_ENV.channels[ch].volume = DEFAULT_VOLUME_LEVEL;
        MIDI_ENV.channels[ch].pan = DEFAULT_PAN_LEVEL;

        MIDI_ENV.channels[ch].chorusProgram = DEFAULT_CHORUS_PROGRAM;
        MIDI_ENV.channels[ch].chorusLevel = 0;
        MIDI_ENV.channels[ch].chorusFeedback = 0;
        MIDI_ENV.channels[ch].chorusDelay = 0;

        MIDI_ENV.channels[ch].reverbProgram = DEFAULT_REVERB_PROGRAM;
        MIDI_ENV.channels[ch].reverbLevel = 0;
        MIDI_ENV.channels[ch].reverbFeedback = 0;

        MIDI_ENV.channels[ch].portamentoState = 0;
        MIDI_ENV.channels[ch].portamentoTime = 0;

        MIDI_ENV.channels[ch].modulationDepth = 0;
        MIDI_ENV.channels[ch].modulationRate = 0x40;
    }
}

/*******************************************************************/
/***             Update Parameter Utility Function               ***/
/*******************************************************************/

// This function is used to inc/dec parameter values 
// Results from update parameter function call.
// RESULT_UPDATE means to update the parameter
// RESULT_CONTINUE means stay in same state
// RESULT_PREV means to return to the previous state
// RESULT_NEXT means to advance to the next state

enum UPDATEPARAMETER_RESULTS {
    RESULT_UPDATE, RESULT_CONTINUE, RESULT_PREVIOUS, RESULT_NEXT
};

// Utility function to update a parameter from the UI
enum UPDATEPARAMETER_RESULTS updateParameter(char *uiLabel, int16_t valueOffset, int16_t *paramPtr,
int16_t minValue, int16_t maxValue, int16_t incr) {

    enum PARAMETER_STATE {
        STATE_SHOW, STATE_UPDATE, STATE_POLLJS, STATE_SET
    };

    static enum PARAMETER_STATE state = STATE_SHOW;
    enum JS_DIRECTION direction;

    switch (state) {
    case STATE_SHOW:
        // Is there a UI label to display ?
        if (strlen(uiLabel) != 0) {
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print(uiLabel);
            lcd.print(*paramPtr + valueOffset);
        }

        state = STATE_UPDATE;
        return RESULT_CONTINUE;

    case STATE_UPDATE:
        state = STATE_POLLJS;
        return RESULT_UPDATE;

    case STATE_POLLJS:
        if (pollRightJoystick()) {
            state = STATE_SET;
        }
        return RESULT_CONTINUE;

    case STATE_SET:
        direction = readRightJoystick();
        switch (direction) {

        case DIR_UP:
            *paramPtr += incr;
            if (*paramPtr > maxValue) {
                *paramPtr = minValue;
            }
            state = STATE_SHOW;
            return RESULT_CONTINUE;

        case DIR_DOWN:
            *paramPtr -= incr;
            if (*paramPtr < minValue) {
                *paramPtr = maxValue;
            }
            state = STATE_SHOW;
            return RESULT_CONTINUE;

        case DIR_RIGHT:
            // Reset state machine to initial state for next time
            state = STATE_SHOW;
            return RESULT_NEXT;

        case DIR_LEFT:
            // Reset state machine to initial state for next time
            state = STATE_SHOW;
            return RESULT_PREVIOUS;
        }
    }
}

/*******************************************************************/
/***               Global (Non-Channel) Functions                ***/
/*******************************************************************/

// Reset synthesizer function
boolean resetFunction(boolean firstTime) {

    enum RESET_STATE {
        STATE_RESET, STATE_DELAY
    };

    static int count;
    static enum RESET_STATE state;

    if (firstTime) {
        state = STATE_RESET;
        return false;
    }

    switch (state) {
    case STATE_RESET:
        // Display status line on lcd
        lcd.setCursor(0, 1);
        lcd.print("Reseting ...");

        // Perform the reset
        midiReset();

        count = LOOP_COUNT;
        state = STATE_DELAY;
        return false;

    case STATE_DELAY:
        count--;
        return (count <= 0) ? true : false;
    }
}

// Select and play a MIDI file
boolean playMidiFileFunction(boolean firstTime) {

    enum PLAY_STATE {
        STATE_GETNAME, STATE_PLAY
    };

    static enum PLAY_STATE state;
    enum JS_DIRECTION direction;
    static char buffer[13];

    if (firstTime) {
        state = STATE_GETNAME;
        return false;
    }

    switch (state) {
    case STATE_GETNAME:
        switch (updateParameter("", 0, &MIDI_ENV.fileIndex, 0, numberOfFiles, 1)) {
        case RESULT_UPDATE:
            // Get the filename at the index
            getMIDIFilename(MIDI_ENV.fileIndex, buffer);

            // Position the cursor and display filename
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print(buffer);
            return false;
        case RESULT_CONTINUE:
            return false;
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_PLAY;
            return false;
        }

    case STATE_PLAY:
        // Indicate which file is playing
        lcd.setCursor(0, 1);
        lcd.print("-> ");
        lcd.print(buffer);

        // Play the MIDI file to completion
        playSMF(buffer);
        return true;
    }
}

/*******************************************************************/
/***            Record/Play Riff Finite State Machines           ***/
/*******************************************************************/

// Calculate required buffer size
#define MAX_RIFF_EVENTS       1000
#define AVG_MIDI_EVENT_LENGTH    7
#define MIDI_BUFFER_SIZE        (MAX_RIFF_EVENTS * AVG_MIDI_EVENT_LENGTH)
#define MAX_RIFF_FILES          20

boolean recording;
int16_t saveFlag;
int16_t riffFileIndex;
int midiDataIndex;
int fileSize;
int midiEventCount;
elapsedMillis midiTicCount;

int midiEventType;
int midiParameter1;
int midiParameter2;

byte midiDataBuffer[MIDI_BUFFER_SIZE];

// Possible states of MIDI recording state machine
enum RECORDING_STATES {
    STATE_GETTYPE, 
    STATE_BYTE2OF3, STATE_BYTE3OF3, 
    STATE_BYTE2OF2,
    STATE_SYSEX,
    STATE_ERROR
};

// Current state of recording state machine
enum RECORDING_STATES recordingState;

// Record a two byte midi event sequence
void recordTwoByteMidiEvent(int midiEventType, int midiParameter) {

    // Is there enough room in the midi data buffer for a new event ?
    if (midiDataIndex < MIDI_BUFFER_SIZE - 6) {
        // Capture current tic count for recording this event
        unsigned long ticCount = midiTicCount;

        midiDataBuffer[midiDataIndex++] = ticCount         & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 8)  & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 16) & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 24) & 0xFF;
        midiDataBuffer[midiDataIndex++] = midiEventType;
        midiDataBuffer[midiDataIndex++] = midiParameter;

        midiEventCount++;
    }
}

// Record a three byte midi event sequence
void recordThreeByteMidiEvent(int midiEventType, int midiParameter1, int midiParameter2) {

    // Is there enough room in the midi data buffer for a new event ?
    if (midiDataIndex < MIDI_BUFFER_SIZE - 7) {
        // Capture current tic count for recording this event
        unsigned long ticCount = midiTicCount;

        midiDataBuffer[midiDataIndex++] = ticCount         & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 8)  & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 16) & 0xFF;
        midiDataBuffer[midiDataIndex++] = (ticCount >> 24) & 0xFF;
        midiDataBuffer[midiDataIndex++] = midiEventType;
        midiDataBuffer[midiDataIndex++] = midiParameter1;
        midiDataBuffer[midiDataIndex++] = midiParameter2;

        midiEventCount++;
    }
}

// Recording Finite State Machine
// Called each time MIDI data is available and we are recording
void runRecordingFSM(int data) {

    switch (recordingState) {
    case STATE_GETTYPE:
        // Save event type
        midiEventType = data;

        switch (midiEventType >> 4) {
        case 0x08:    // Note off
        case 0x09:    // Note on
        case 0x0A:    // Note aftertouch
        case 0x0B:    // Controller change
        case 0x0E:    // Pitch bend
            recordingState = STATE_BYTE2OF3;
            break;

        case 0x0C:    // Program change
        case 0x0D:    // Channel aftertouch
            recordingState = STATE_BYTE2OF2;
            break;

        case 0x0F:
            recordingState = STATE_SYSEX;
            break;

        default:
            recordingState = STATE_ERROR;
            break;
        }
        break;

    case STATE_BYTE2OF3:
        midiParameter1 = data;
        recordingState = STATE_BYTE3OF3;
        break;

    case STATE_BYTE3OF3:
        midiParameter2 = data;
        // Record a 3 byte midi event
        recordThreeByteMidiEvent(midiEventType, midiParameter1, midiParameter2);
        recordingState = STATE_GETTYPE;
        break;

    case STATE_BYTE2OF2:
        midiParameter1 = data;
        // Record a 2 byte midi event
        recordTwoByteMidiEvent(midiEventType, midiParameter1);
        recordingState = STATE_GETTYPE;
        break;

    case STATE_SYSEX:
        // Consume all of the sysex message
        if (data == 0xF7) {
            recordingState = STATE_GETTYPE;
        }
        break;

    case STATE_ERROR:
        while (true) {
            Serial.println("MIDI out of sync");
            delay(500);
        }
        break;
    }
}

// Record a MIDI riff function
boolean recordRiffFunction(boolean firstTime) {

    enum RECORDING_UI_STATE {
        STATE_RECORD_START, STATE_POLLJS, STATE_RECORD_STOP, STATE_GETSAVE, STATE_GETFILE
    };

    static enum RECORDING_UI_STATE recordingUIState;
    enum JS_DIRECTION direction;

    if (firstTime) {
        // Do some initialization
        recording = false;    
        recordingState = STATE_GETTYPE;
        saveFlag = 0;
        riffFileIndex = 1;

        // Clear data buffer in preparation for recording
        midiDataIndex = 0;
        midiEventCount = 0;
        memset(midiDataBuffer, 0, sizeof(midiDataBuffer));

        // Set initial state
        recordingUIState = STATE_RECORD_START;
        return false;
    }

    switch (recordingUIState) {
    case STATE_RECORD_START:
        lcd.setCursor(0, 1);
        lcd.print("                ");
        lcd.setCursor(0, 1);
        lcd.print("Recording Riff");

        // Indicate we are recording
        recording = true;

        // Reset the tic count
        midiTicCount = 0;

        // Set next state
        recordingUIState = STATE_POLLJS;
        return false;

    case STATE_POLLJS:
        if (pollRightJoystick()) {
            recordingUIState = STATE_RECORD_STOP;
        }
        return false;

    case STATE_RECORD_STOP:
        // Stop recording
        recording = false;

        direction = readRightJoystick();
        switch (direction) {
        case DIR_RIGHT:
            if (midiEventCount != 0) {
                // Set next state
                recordingUIState = STATE_GETSAVE;
                return false;
            }    
            else    {
                // If no midi events, we're done
                return true;
            }

        case DIR_UP:
        case DIR_DOWN:
        case DIR_LEFT:
            // We're done
            return true;
        }
    case STATE_GETSAVE:
        switch (updateParameter("", 0, &saveFlag, 0, 1, 1)) {
        case RESULT_UPDATE:
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Save Riff: ");
            lcd.print((saveFlag == 1) ? "Yes" : "No");
            return false;
        case RESULT_PREVIOUS:
            // We're done
            return true;
        case RESULT_NEXT:
            if (saveFlag == 1) {
                recordingUIState = STATE_GETFILE;
                return false;
            }    
            else    {
                // We're done
                return true;
            }
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETFILE:
        switch (updateParameter("File: RIFF", 0, &riffFileIndex, 1, MAX_RIFF_FILES, 1)) {
        case RESULT_UPDATE:
            return false;
        case RESULT_PREVIOUS:
            // We're done
            return true;
        case RESULT_NEXT:
            {
                // Save the riff in a riff file
                lcd.setCursor(0, 1);
                lcd.print("                ");
                lcd.setCursor(0, 1);
                lcd.print("Saving RIFF file");

                SdFile riffFile;
                char filename[15];
                char temp[6];

                // Format riff filename
                memset(filename, 0, sizeof(filename));
                strcpy(filename, "RIFF");
                itoa(riffFileIndex, temp, 10);
                strcat(filename, temp);
                strcat(filename, ".rif");

                // Open the riff file for writing
                if (! riffFile.open(filename, O_CREAT | O_WRITE)) {
                    lcd.clear();
                    lcd.setCursor(0, 0);
                    lcd.print("Error - Creating");
                    lcd.setCursor(0, 1);
                    lcd.print("    RIFF file   ");
                    delay(2000);
                    // No more we can do. We're done
                    return true;
                }

                // Write the number of events to the file
                riffFile.write(midiEventCount & 0xFF);
                riffFile.write((midiEventCount >> 8) & 0xFF);

                // Write the data
                riffFile.write(midiDataBuffer, midiDataIndex);

                // Close the file
                riffFile.close();
            }
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}

// Play a recorded MIDI riff function
boolean playRiffFunction(boolean firstTime) {

    enum PLAYBACK_UI_STATE {
        STATE_GETFILE, STATE_PLAY
    };

    static enum PLAYBACK_UI_STATE playbackUIState;
    enum JS_DIRECTION direction;

    if (firstTime) {
        // Do some initialization
        recording = false;    
        riffFileIndex = 1;

        // Clear data buffer in preparation for playback
        midiDataIndex = 0;
        memset(midiDataBuffer, 0, sizeof(midiDataBuffer));

        // Set initial state
        playbackUIState = STATE_GETFILE;
        return false;
    }

    switch (playbackUIState) {
    case STATE_GETFILE:
        switch (updateParameter("File: RIFF", 0, &riffFileIndex, 1, MAX_RIFF_FILES, 1)) {
        case RESULT_UPDATE:
            return false;
        case RESULT_PREVIOUS:
            // We're done
            return true;
        case RESULT_NEXT:
            {
                SdFile riffFile;
                char filename[15];
                char temp[6];

                // Format riff filename
                memset(filename, 0, sizeof(filename));
                strcpy(filename, "RIFF");
                itoa(riffFileIndex, temp, 10);
                strcat(filename, temp);
                strcat(filename, ".rif");

                // Open the riff file for reading
                if (! riffFile.open(filename, O_READ)) {
                    lcd.clear();
                    lcd.setCursor(0, 0);
                    lcd.print("Error - No such ");
                    lcd.setCursor(0, 1);
                    lcd.print("    RIFF file   ");
                    delay(2000);
                    // No more we can do. We're done
                    return true;
                }
                // Get the size of riff in bytes
                fileSize = riffFile.fileSize();

                // Get the event count
                midiEventCount  = riffFile.read();
                midiEventCount |= (riffFile.read() << 8);

                // Read the data into the buffer
                riffFile.read(midiDataBuffer, fileSize - 2);

                // Close the file
                riffFile.close();

                playbackUIState = STATE_PLAY;
            }
            return false;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_PLAY:
        {
            unsigned long ticCount;
            int eventType;
            int channel;

            // Play the riff in the midiDataBuffer
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Playing Riff");

            // Reset the timebase
            midiTicCount = 0;

            // While there is data to process and user has not touched the right joystick
            while ((midiEventCount--) && (! pollRightJoystick())) {
                // Get the MIDI event tic count
                ticCount =  (((unsigned long) midiDataBuffer[midiDataIndex++]));
                ticCount |= (((unsigned long) midiDataBuffer[midiDataIndex++]) << 8);
                ticCount |= (((unsigned long) midiDataBuffer[midiDataIndex++]) << 16);
                ticCount |= (((unsigned long) midiDataBuffer[midiDataIndex++]) << 24);

                midiEventType  = midiDataBuffer[midiDataIndex++];
                midiParameter1 = midiDataBuffer[midiDataIndex++];
                midiParameter2 = midiDataBuffer[midiDataIndex++];

                eventType = midiEventType >> 4;
                channel   = midiEventType & 0x0F;

                // Wait for the event time to continue
                while (ticCount >= midiTicCount) {
                    ;
                }
                // Now process the midi codes
                switch (eventType) {
                case 0x08:    // Note off
                    noteOff(channel, midiParameter1);
                    break;

                case 0x09:    // Note on
                    noteOn(channel, midiParameter1, midiParameter2);
                    break;

                case 0x0A:    // Note aftertouch
                    // Ignore
                    break;

                case 0x0B:    // Control change
                    controllerChange(channel, midiParameter1, midiParameter2);
                    break;

                case 0x0C:    // Program change
                    programChange(0x00, channel, midiParameter1);
                    midiDataIndex--;
                    break;

                case 0x0D:    // Channel aftertouch
                    // Ignore
                    midiDataIndex--;
                    break;

                case 0x0E:    // Pitch bend
                    pitchBend(channel, midiParameter1);
                    break;
                }
            }
        }
        return true;
    }
}


// Select the drum set for MIDI channel 10
// Sets are: the Standard set, the Power set, the Brush set, the Orchestral set and the CM-64/32 set 
boolean drumSetFunction(boolean firstTime) {

    switch (updateParameter("", 0, &MIDI_ENV.drumSetIndex, 0, NUMBER_OF_DRUMSETS - 1, 1)) {
    case RESULT_UPDATE:
        lcd.setCursor(0, 1);
        lcd.print("                ");
        lcd.setCursor(0, 1);
        lcd.print(DRUM_TYPES[MIDI_ENV.drumSetIndex].name);

        // Do the change so results are immediately available
        drumSetChange(DRUM_TYPES[MIDI_ENV.drumSetIndex].value);
        return false;
    case RESULT_CONTINUE:
        return false;
    case RESULT_PREVIOUS:
    case RESULT_NEXT:
        // We're done
        return true;
    }
}

// Set the master volume
boolean masterVolumeFunction(boolean firstTime) {

    switch (updateParameter("Volume: ", 0, &(MIDI_ENV.masterVolume), MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
    case RESULT_UPDATE:
        // Set the currently selected volume
        setMasterVolume(MIDI_ENV.masterVolume);
    case RESULT_CONTINUE:
        return false;
    case RESULT_PREVIOUS:
    case RESULT_NEXT:
        return true;
    }
}

/*******************************************************************/
/***                  Channel Based Functions                    ***/
/*******************************************************************/

// Set channel volume
boolean channelVolumeFunction(boolean firstTime) {

    enum VOLUME_STATE {
        STATE_GETCHANNEL, STATE_GETVOLUME
    };

    static int16_t channel;
    static enum VOLUME_STATE state;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_GETVOLUME;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETVOLUME:
        switch (updateParameter("Volume: ", 0, &(MIDI_ENV.channels[channel].volume), MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Set the currently selected volume
            setChannelVolume(channel, MIDI_ENV.channels[channel].volume);
        case RESULT_CONTINUE:
            return false;
        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            return true;
        }
    }
}

// Set channel pan
boolean channelPanFunction(boolean firstTime) {

    enum PAN_STATE {
        STATE_GETCHANNEL, STATE_GETPAN
    };

    static int16_t channel;
    static enum PAN_STATE state;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            // We're done
            return true;
        case RESULT_NEXT:
            state = STATE_GETPAN;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETPAN:
        switch (updateParameter("Pan: ", 0, &(MIDI_ENV.channels[channel].pan), 0, 127, 4)) {
        case RESULT_UPDATE:
            // Set the currently selected volume
            setChannelPan(channel, MIDI_ENV.channels[channel].pan);
        case RESULT_CONTINUE:
            return false;
        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            // We're done
            return true;
        }
    }
}

// Channel program selection
boolean channelProgramFunction(boolean firstTime) {

    enum PC_STATE {
        STATE_GETCHANNEL, STATE_GETBANK, STATE_GETPROGRAM 
    };

    static enum PC_STATE state;
    static int16_t channel;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_GETBANK;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETBANK:
        switch (updateParameter("", 0, &MIDI_ENV.channels[channel].bank, 0, 1, 1)) {
        case RESULT_UPDATE:
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Bank: ");
            lcd.print((MIDI_ENV.channels[channel].bank == 0) ? "A" : "B");
            return false;
        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETPROGRAM;
            return false;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETPROGRAM:
        switch (updateParameter("", 0, &MIDI_ENV.channels[channel].program, 0, NUMBER_OF_BANK_VOICES - 1, 1)) {
        case RESULT_UPDATE:
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);

            // Show the current program
            lcd.print("Prg: ");
            if (MIDI_ENV.channels[channel].bank == 0) {
                lcd.print(BANK0_VOICES[MIDI_ENV.channels[channel].program]);
            }    
            else    {
                lcd.print(BANK1_VOICES[MIDI_ENV.channels[channel].program]);
            }
            // Do the program change so results are heard immediately
            programChange ((MIDI_ENV.channels[channel].bank == 0) ? 0x00 : 0x7F, channel, MIDI_ENV.channels[channel].program);
            return false;
        case RESULT_PREVIOUS:
            state = STATE_GETBANK;
            return false;
        case RESULT_NEXT:
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}

// Channel Chorus Effect Function
boolean channelChorusFunction(boolean firstTime) {

    enum CHORUS_STATE {
        STATE_GETCHANNEL, STATE_GETLEVEL, STATE_GETPROGRAM, STATE_GETFEEDBACK, STATE_GETDELAY
    };

    static enum CHORUS_STATE state;
    static int16_t channel;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_GETLEVEL;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETLEVEL:
        switch (updateParameter("Level: ", 0, &MIDI_ENV.channels[channel].chorusLevel, 
        MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setChorus(channel, MIDI_ENV.channels[channel].chorusProgram, MIDI_ENV.channels[channel].chorusLevel, 
            MIDI_ENV.channels[channel].chorusFeedback, MIDI_ENV.channels[channel].chorusDelay);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETPROGRAM;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETPROGRAM:
        switch (updateParameter("", 0, &MIDI_ENV.channels[channel].chorusProgram, 0, NUMBER_OF_CHORUSES - 1, 1)) {
        case RESULT_UPDATE:
            // Show the current program
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Chorus: ");
            lcd.print(CHORUS_TYPES[MIDI_ENV.channels[channel].chorusProgram]);

            // Do the change so results are heard immediately
            setChorus(channel, MIDI_ENV.channels[channel].chorusProgram, MIDI_ENV.channels[channel].chorusLevel, 
            MIDI_ENV.channels[channel].chorusFeedback, MIDI_ENV.channels[channel].chorusDelay);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETLEVEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETFEEDBACK;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETFEEDBACK:
        switch (updateParameter("Feedback: ", 0, &MIDI_ENV.channels[channel].chorusFeedback, MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setChorus(channel, MIDI_ENV.channels[channel].chorusProgram, MIDI_ENV.channels[channel].chorusLevel, 
            MIDI_ENV.channels[channel].chorusFeedback, MIDI_ENV.channels[channel].chorusDelay);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETPROGRAM;
            return false;
        case RESULT_NEXT:
            state = STATE_GETDELAY;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETDELAY:
        switch (updateParameter("Delay: ", 0, &MIDI_ENV.channels[channel].chorusDelay, MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setChorus(channel, MIDI_ENV.channels[channel].chorusProgram, MIDI_ENV.channels[channel].chorusLevel, 
            MIDI_ENV.channels[channel].chorusFeedback, MIDI_ENV.channels[channel].chorusDelay);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETFEEDBACK;
            return false;
        case RESULT_NEXT:
            // We're done
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}

// Channel Reverb Effect Function
boolean channelReverbFunction(boolean firstTime) {

    enum REVERB_STATE {
        STATE_GETCHANNEL, STATE_GETLEVEL, STATE_GETPROGRAM, STATE_GETFEEDBACK
    };

    static enum REVERB_STATE state;
    static int16_t channel;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_GETLEVEL;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETLEVEL:
        switch (updateParameter("Level: ", 0, &MIDI_ENV.channels[channel].reverbLevel, 
        MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setReverb(channel, MIDI_ENV.channels[channel].reverbProgram, MIDI_ENV.channels[channel].reverbLevel, 
            MIDI_ENV.channels[channel].reverbFeedback);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETPROGRAM;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETPROGRAM:
        switch (updateParameter("", 0, &MIDI_ENV.channels[channel].reverbProgram, 0, NUMBER_OF_REVERBS - 1, 1)) {
        case RESULT_UPDATE:
            // Show the current program
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Reverb: ");
            lcd.print(REVERB_TYPES[MIDI_ENV.channels[channel].reverbProgram]);

            // Do the change so results are heard immediately
            setReverb(channel, MIDI_ENV.channels[channel].reverbProgram, MIDI_ENV.channels[channel].reverbLevel, 
            MIDI_ENV.channels[channel].reverbFeedback);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETLEVEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETFEEDBACK;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETFEEDBACK:
        switch (updateParameter("Feedback: ", 0, &MIDI_ENV.channels[channel].reverbFeedback, MIDI_MIN_VOLUME, MIDI_MAX_VOLUME, VOLUME_INCREMENT)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setReverb(channel, MIDI_ENV.channels[channel].reverbProgram, MIDI_ENV.channels[channel].reverbLevel, 
            MIDI_ENV.channels[channel].reverbFeedback);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETPROGRAM;
            return false;
        case RESULT_NEXT:
            // We're done
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}

// Portamento function
boolean channelPortamentoFunction(boolean firstTime) {

    enum PORTAMENTO_STATE {
        STATE_GETCHANNEL, STATE_GETSTATE, STATE_GETTIME
    };

    static enum PORTAMENTO_STATE state;
    static int16_t channel;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            return true;
        case RESULT_NEXT:
            state = STATE_GETSTATE;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETSTATE:
        switch (updateParameter("", 0, &MIDI_ENV.channels[channel].portamentoState, 0, 1, 1)) {
        case RESULT_UPDATE:
            lcd.setCursor(0, 1);
            lcd.print("                ");
            lcd.setCursor(0, 1);
            lcd.print("Portamento: ");
            lcd.print((MIDI_ENV.channels[channel].portamentoState == 0) ? "Off" : "On");

            // Do the change so results are heard immediately
            setPortamento(channel, MIDI_ENV.channels[channel].portamentoState, MIDI_ENV.channels[channel].portamentoTime); 
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETTIME;
            return false;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETTIME:
        switch (updateParameter("Time: ", 0, &MIDI_ENV.channels[channel].portamentoTime, 0, 127, 4)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setPortamento(channel, MIDI_ENV.channels[channel].portamentoState, MIDI_ENV.channels[channel].portamentoTime); 
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETSTATE;
            return false;
        case RESULT_NEXT:
            // We're done
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}

// Modulation function
boolean channelModulationFunction(boolean firstTime) {

    enum MODULATION_STATE {
        STATE_GETCHANNEL, STATE_GETDEPTH, STATE_GETRATE
    };

    static enum MODULATION_STATE state;
    static int16_t channel;
    enum JS_DIRECTION direction;

    if (firstTime) {
        state = STATE_GETCHANNEL;
        channel = 0;
        return false;
    }

    switch (state) {
    case STATE_GETCHANNEL:
        switch (updateParameter("Channel: ", 1, &channel, 0, 15, 1)) {
        case RESULT_PREVIOUS:
            // We're done
            return true;
        case RESULT_NEXT:
            state = STATE_GETDEPTH;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETDEPTH:
        switch (updateParameter("Depth: ", 0, &MIDI_ENV.channels[channel].modulationDepth, 0, 127, 4)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setModulation(channel, MIDI_ENV.channels[channel].modulationDepth, MIDI_ENV.channels[channel].modulationRate);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETCHANNEL;
            return false;
        case RESULT_NEXT:
            state = STATE_GETRATE;
            return false;
        case RESULT_CONTINUE:
            return false;
        }

    case STATE_GETRATE:
        // NOTE: rate is global across all channels !
        switch (updateParameter("Rate: ", 0, &MIDI_ENV.channels[channel].modulationRate, 0, 127, 4)) {
        case RESULT_UPDATE:
            // Do the change so results are heard immediately
            setModulation(channel, MIDI_ENV.channels[channel].modulationDepth, MIDI_ENV.channels[channel].modulationRate);
            return false;

        case RESULT_PREVIOUS:
            state = STATE_GETDEPTH;
            return false;
        case RESULT_NEXT:
            // We're done
            return true;
        case RESULT_CONTINUE:
            return false;
        }
    }
}









